/* $Id: mle.hpp 158 2010-09-13 22:02:22Z pjohnson $
  ---------------------------------------------------------------------------
  This file is part of the PIIM package; please see:
    Johnson, PLF and Slatkin, M. 2009. "Inference of Microbial
    Recombination Rates from Metagenomic Data."  PLoS Genetics (in
    press).
 
  Copyright 2009 The Regents of the University of California.  All
  rights reserved.
 
  IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
  INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
  LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
  DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY
  OF SUCH DAMAGE.
 
  REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING
  DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
  IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
  UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 
  This software is distributed under the terms of Version 2 of the
  GNU General Public License as published by the Free Software
  Foundation.  If you have not received a copy of Version 2 of the
  GNU General Public License with this program, you may download the
  GPL from the following url: http://www.gnu.org/copyleft/gpl.html.

  ---------------------------------------------------------------------------
  Generic maximum likelihood estimation classes
  ---------------------------------------------------------------------------
*/

#include "gsl_util.hpp"
#include "util.hpp"
#include <vector>
#include <stdexcept>


/*---------------------------------------------------------------------------*/
class CParams {
public:
    typedef std::vector<double> TParams;

public:
    CParams(unsigned int numParams) : m_Params(numParams) {}

    double Get(unsigned int p) const { return m_Params[p]; }
    double operator[](unsigned int p) const { return m_Params[p]; }
    CParams& Set(unsigned int p, double d) { m_Params[p] = d; return *this;}
    double& Set(unsigned int p) { return m_Params[p]; }

    bool operator< (const CParams &p) const {
        return m_Params < p.m_Params;
    }
    friend std::ostream& operator<< (std::ostream& o, const CParams &p) {
        for (unsigned int i = 0;  i < p.m_Params.size();  ++i) {
            if (i != 0) {
                o << "\t";
            }
            o << p.m_Params[i];
        }
        return o;
    }

protected:
    TParams m_Params;
};


/*---------------------------------------------------------------------------*/

class CLikelihoodModel {
public:
    virtual ~CLikelihoodModel(void) {} //silence compiler warning
    virtual double CalcLogLik(const CParams &par) const = 0;
};

/*---------------------------------------------------------------------------*/

class CFindMLE : gsl::CSimplex {
public:
    typedef std::vector<int> TParamNames;

public:
    CFindMLE(const CLikelihoodModel &model) :
        m_Model(model), m_Params(0), m_MinBound(0), m_MaxBound(0),
        m_Debug(false), m_Tolerance(0.0001), m_StepSize(0.01) {}

    CFindMLE& SetTolerance(double t) { m_Tolerance = t; return *this; }
    CFindMLE& SetStepSize(double s) { m_StepSize = s; return *this; }
    CFindMLE& SetDebug(bool d = true) { m_Debug = d; return *this; }

    void Maximize(const CParams &p, const TParamNames &names,
                  int profileParam = -1) {
        m_Params = m_MinBound = m_MaxBound = p;
        m_WhichParams = names;
        m_ProfileParam = profileParam;
        //only optimize if we have something to optimize
        if (m_WhichParams.size() > 0  &&
            (m_WhichParams.size() != 1  ||  m_ProfileParam == -1)) {
            Optimize(m_StepSize, m_Tolerance, 300);
        }
    }
    
    //convenience constructor
    CFindMLE(const CLikelihoodModel &model, const CParams &p,
             const TParamNames &names, double tolerance=0.0001,
             int profileParam=-1, bool debug = false) :
        m_Model(model), m_Params(0), m_MinBound(0), m_MaxBound(0)
    {
        SetTolerance(tolerance).SetStepSize(0.01).SetDebug(debug);
        Maximize(p, names, profileParam);
    }
    const CParams& GetMLParams(void) const { return m_Params; }
    const CParams& GetMaxExploredParams(void) const { return m_MaxBound; }
    const CParams& GetMinExploredParams(void) const { return m_MinBound; }
    bool IsMLatBoundary(void) const {
        for (TParamNames::const_iterator p = m_WhichParams.begin();
             p != m_WhichParams.end();  ++p) {
            if (m_ProfileParam != *p) {
                if (m_MinBound.Get(*p) == m_Params.Get(*p)  ||
                    m_MaxBound.Get(*p) == m_Params.Get(*p)) {
                    return true;
                }
            }
        }
        return false;
    }

protected:
    virtual double F(const gsl_vector* x) const {
        try {
            CParams p(m_Params);
            x_Vector2Params(x, p);
            double r = -m_Model.CalcLogLik(p);
            if (m_Debug) {
                std::cerr << "F: " << p << "\t" << " --> " << r << std::endl;
            }
            
            for (TParamNames::const_iterator i = m_WhichParams.begin();
                 i != m_WhichParams.end();  ++i) {
                if (m_ProfileParam != *i) {
                    if (m_MinBound.Get(*i) > p.Get(*i)) {
                        m_MinBound.Set(*i) = p.Get(*i);
                    } else if (m_MaxBound.Get(*i) < p.Get(*i)) {
                        m_MaxBound.Set(*i) = p.Get(*i);
                    }
                }
            }
            return r;
        } catch (std::range_error&) { // parameter out of range
            return -log(0);
        }
    }

    virtual unsigned int x_GetNumParams(void) const {
        return (m_ProfileParam == -1 ?
                m_WhichParams.size():m_WhichParams.size()-1);
    }
    virtual void x_InitializeParams(gsl_vector* x) const {
        int i = 0;
        for (TParamNames::const_iterator p = m_WhichParams.begin();
             p != m_WhichParams.end();  ++p) {
            if (m_ProfileParam != *p) {
                gsl_vector_set(x, i++, m_Params.Get(*p));
            }
        }
    }
    virtual void x_StoreParams(const gsl_vector* x) {
        x_Vector2Params(x, m_Params);
    }
    void x_Vector2Params(const gsl_vector* x, CParams &par) const {
        int i = 0;

        for (TParamNames::const_iterator p = m_WhichParams.begin();
             p != m_WhichParams.end();  ++p) {
            if (m_ProfileParam != *p) {
                par.Set(*p, gsl_vector_get(x, i++));
            }
        }
    }

private:
    const CLikelihoodModel &m_Model;
    CParams m_Params;
    //min & max explored *legit* parameters
    mutable CParams m_MinBound, m_MaxBound;
    TParamNames m_WhichParams;
    int m_ProfileParam;
    bool m_Debug;
    double m_Tolerance;
    double m_StepSize;
};

/*---------------------------------------------------------------------------*/

class CFindCI : gsl::CFindRootBracket {
public:
    typedef std::vector<int> TParamNames;
public:
    CFindCI(const CLikelihoodModel &m, const CParams &mle,
            const TParamNames &whichParams, CParams &lbPar, CParams &ubPar,
            double tolerance = 0.0001)
        : m_Model(m), m_MLE(mle), m_ML(m.CalcLogLik(mle)),
          m_WhichParams(whichParams), m_Tolerance(tolerance) {
        const int kMaxIter = 100;
        m_ParamCache.Set(m_MLE) = m_ML;

        for (TParamNames::const_iterator p = m_WhichParams.begin();
             p != m_WhichParams.end();  ++p) {
            m_Which = *p;
            double lb, ub;
            x_SetSearchRange(lb, ub);
            if (lb > ub) { //for negative parameters
                swap(lb, ub);
            }

            CParams par(m_MLE);
            {//lower c.i.
                par.Set(m_Which, lb);
                CFindMLE profileMax(m_Model, par, m_WhichParams, m_Tolerance,
                                    m_Which);
                double llik = m_Model.CalcLogLik(profileMax.GetMLParams());
                if (llik > m_ML-2.0) {
                    lbPar.Set(m_Which, lb); //FIXME: need to flag < lb
                } else {
                    m_ParamCache.Set(par) = llik;
                    lbPar.Set(m_Which,
                              FindRoot(lb, mle.Get(m_Which), m_Tolerance,
                                       kMaxIter));
                }
            }
            {//upper c.i.
                par.Set(m_Which, ub);
                CFindMLE profileMax(m_Model, par, m_WhichParams, m_Tolerance,
                                    m_Which);
                double llik = m_Model.CalcLogLik(profileMax.GetMLParams());
                if (llik > m_ML-2.0) {
                    ubPar.Set(m_Which, ub); //FIXME: need to flag > ub
                } else {
                    m_ParamCache.Set(par) = llik;
                    ubPar.Set(m_Which,
                              FindRoot(mle.Get(m_Which), ub, m_Tolerance,
                                       kMaxIter));
                }
            }
        }
    }
protected:
    //overide this in derived class if want to tweak search range
    virtual void x_SetSearchRange(double &lb, double &ub) const {
        lb = m_MLE.Get(m_Which)/10;
        ub = min(0.99, m_MLE.Get(m_Which)*2);
    }

private:
    virtual double F(const double s) const {
        CParams par(m_MLE);
        par.Set(m_Which, s);

        double llik;
        try {
            llik = m_ParamCache[par];
        } catch (...) { //missed cache
            CFindMLE profileMax(m_Model, par, m_WhichParams, m_Tolerance,
                                m_Which);
            llik = m_Model.CalcLogLik(profileMax.GetMLParams());
        }

        double r = llik - (m_ML-2.0);
        //cerr << "CF: " << par << "\t" << " --> " << r << endl;
        return r;
    }
    
    const CLikelihoodModel &m_Model;
    const CParams &m_MLE;
    double m_ML;
    CCache<double, const CParams> m_ParamCache;
    const TParamNames &m_WhichParams;
    unsigned int m_Which;
    double m_Tolerance;
};

