/* $Id: piim_data.hpp 156 2010-09-13 21:15:29Z pjohnson $
  ---------------------------------------------------------------------------
  This file is part of the PIIM package; please see:
    Johnson, PLF and Slatkin, M. 2009. "Inference of Microbial
    Recombination Rates from Metagenomic Data."  PLoS Genetics (in
    press).
 
  Copyright 2009 The Regents of the University of California.  All
  rights reserved.
 
  IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
  INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
  LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
  DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY
  OF SUCH DAMAGE.
 
  REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING
  DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
  IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
  UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 
  This software is distributed under the terms of Version 2 of the
  GNU General Public License as published by the Free Software
  Foundation.  If you have not received a copy of Version 2 of the
  GNU General Public License with this program, you may download the
  GPL from the following url: http://www.gnu.org/copyleft/gpl.html.

  ---------------------------------------------------------------------------
  Classes for handling metagenomic data with PIIM
  ---------------------------------------------------------------------------
*/
#ifndef PIIM_DATA_HPP
#define PIIM_DATA_HPP

#include <vector>
#include <iostream>
#include <stdexcept>
#include <cassert>
#include <algorithm>

enum ENuc {
    eA=0, eC, eG, eT, eN, eGap
};

/*---------------------------------------------------------------------------*/
// PRE : nucleotide enum
// POST: string representation
inline char Enum2Ch(ENuc nuc) {
    switch (nuc) {
    case eA: return  'A';
    case eC: return  'C';
    case eG: return  'G';
    case eT: return  'T';
    case eGap: return  '-';
    case eN: return  'N';
    default: assert(false);
    }
}

/*---------------------------------------------------------------------------*/
// PRE : character representation
// POST: nucleotide enum
inline ENuc Ch2Enum(char c) {
    switch (c) {
    case 'A': return eA;
    case 'C': return eC;
    case 'G': return eG;
    case 'T': return eT;
    case '-': return eGap;
    case 'X':
    case 'N': return eN;
    default: throw std::runtime_error(std::string("Unknown nucleotide symbol '")+c+"'");
    }
}

//-------------------------------------------------------------------
class CChr; //forward def used in CRead to point to containing chromosome

class CRead {
public:
    typedef std::vector<ENuc> TBases;
    typedef std::vector<unsigned short int> TQuals;

public:
    CRead(unsigned int start = 0) : m_Start(start) {}
    void SetAcc(const std::string& acc) { m_Acc = acc; }
    void SetStart(unsigned int start) { m_Start = start; }
    TBases& SetBases(void) { return m_Bases; }
    TQuals& SetQuals(void) { return m_Quals; }
    void SetChr(const CChr* chr) { m_Chr = chr; }

    //getters
    ENuc operator[] (unsigned int i) const { return GetBase(i); }
    ENuc GetBase(unsigned int i) const {
        if (i >= m_Start  &&  i < m_Start + m_Bases.size()) {
            return m_Bases[i - m_Start];
        } else {
            return eGap;
        }
    }
    std::string GetBases(void) const {//should only be used for debugging
        std::string s(GetLength(), '?');
        for (unsigned int i = GetStart();  i <= GetStop();  ++i) {
            s[i-GetStart()] = Enum2Ch(GetBase(i));
        };
        return s;
    }
    int GetQual(unsigned int i) const { return m_Quals[i - m_Start]; }
    const CChr* GetChr(void) const { return m_Chr; }

    const std::string& GetAcc(void) const { return m_Acc; }
    unsigned int GetStart(void) const { return m_Start; }
    unsigned int GetStop(void) const { return m_Start + m_Bases.size() - 1; }
    unsigned int GetLength(void) const { return m_Bases.size(); }
    bool Overlaps(unsigned int p) const {
        return GetStart() <= p  &&  p <= GetStop(); }

    bool operator< (const CRead &r) const {
        return GetStart() < r.GetStart();
    }
    class CPtrCmp {
    public: bool operator() (const CRead *a, const CRead *b) { return *a<*b; }
    };

private:
    std::string m_Acc;
    unsigned int m_Start;
    TBases m_Bases;
    TQuals m_Quals;
    const CChr *m_Chr;
};

//-------------------------------------------------------------------

class CChr {
public:
    typedef std::vector<CRead> TReads;

public:
    //setters
    TReads& SetReads(void) { return m_Reads; }
    void AddRead(const CRead& r) {
        m_Reads.push_back(r);
        m_Reads.back().SetChr(this);
        inplace_merge(m_Reads.begin(), m_Reads.end()-1, m_Reads.end());
    }

    //getters
    const TReads& GetReads(void) const { return m_Reads; }
    unsigned int GetStart(void) const {
        assert(!m_Reads.empty());
        return m_Reads.front().GetStart();
    }
    unsigned int GetStop(void) const {
        assert(!m_Reads.empty());
        return m_Reads.back().GetStop();
    }

    bool operator< (const CChr &c) const {
        return GetStart() < c.GetStart();
    }
    class CPtrCmp {
    public: bool operator() (const CChr *a, const CChr *b) { return *a < *b; }
    };

private:
    TReads m_Reads;
};

//-------------------------------------------------------------------

class CAssembly {
public:
    typedef std::vector<CChr*> TChrs;
    typedef TChrs::const_iterator const_iterator;

public:
    CAssembly(void) : m_StopPos(-1) {}
    ~CAssembly(void) {
        for (TChrs::iterator i = m_Chrs.begin();  i != m_Chrs.end();  ++i) {
            delete *i;
        }
    }

    void Save(std::ostream&) const;
    void Load(std::istream&);

    void AddChr(CChr *c) {
        m_Chrs.push_back(c);
        inplace_merge(m_Chrs.begin(), m_Chrs.end()-1, m_Chrs.end(),
                      CChr::CPtrCmp());
        m_StopPos = -1; //reset
    }

    //getters
    const CChr& operator[] (unsigned int i) const { return *m_Chrs[i]; }
    const_iterator begin(void) const { return m_Chrs.begin(); }
    const_iterator end(void) const { return m_Chrs.end(); }

    unsigned int GetNumChrs(void) const { return m_Chrs.size(); }
    unsigned int GetStart(void) const {
        return (m_Chrs.empty() ? 0 : m_Chrs[0]->GetStart()); }
    unsigned int GetStop(void) const { 
        if (m_StopPos == -1) {
            x_CacheStop();
        }
        return m_StopPos;
    }
    
private:
    void x_CacheStop(void) const {
        m_StopPos = 0;
        for (TChrs::const_reverse_iterator it = m_Chrs.rbegin();
             it != m_Chrs.rend();  ++it) {
            if ((*it)->GetStop() > (unsigned int) m_StopPos) {
                m_StopPos = (*it)->GetStop();
            }
        }
    }

private:
    TChrs m_Chrs;
    mutable int m_StopPos; //used only for caching stop position
};

#endif // PIIM_DATA_HPP
